
require 'forwardable'
require 'childprocess'
require 'colorize'

ChildProcess.posix_spawn = true

module LTools
  autoload :Shellwords, 'shellwords'
  autoload :Tempfile,   'tempfile'

  class Tool
    extend Forwardable

    attr_reader :stdin, :stdout, :stderr, :cmd
    attr_accessor :verbose

    def_delegators :@process, :stop, :pid, :exited?, \
                   :alive?, :exit_code, :poll_for_exit

    def initialize(*args)
      tool, *args = args

      # 检测参数类型进行变换，支持一下格式:
      #   args = 'echo 1231'
      #   args = ['echo', '1231']
      #   args = [['echo', '1231']]
      case tool
      when String
        if args.empty? and tool.include? ' '
          tool, *args = Shellwords.split(tool)
        end
      when Array
        tool, *args = tool
      else
        raise
      end

      @cmd = args.unshift(tool).map(&:to_s)
      @process = ChildProcess.build(*@cmd)

      @stdout, @stdout_w = IO.pipe
      @process.io.stdout = @stdout_w 

      @stderr, @stderr_w = IO.pipe
      @process.io.stderr = @stderr_w

      @verbose = true if LTools.verbose
    end

    def start
      @process.duplex    = true unless @inherit
      @process.start
      @stdin = @process.io.stdin unless @inherit
      if @verbose
        puts "[#{'+'.light_green.bold}] #{self.class.to_s.light_yellow} (pid: #{pid})"
        puts "  #{?$.light_red.bold} #{command}"
        puts
      end
    end

    def wait
      @process.wait
      @stdout_w.close
      @stderr_w.close
    end

    def stop
      @process.stop
      @stdout_w.close
      @stderr_w.close
    end

    def command
      @cmd.map{|str|
        if str.include?(' ')
          if str.include? '"'
            "'#{str}'"
          else
            "\"#{str}\""
          end
        else
          str
        end
      }.join ' '
    end

    def io_inherit!
      @inherit = true
      @process.io.inherit!
    end

    def self.inherited(child_class)
      LTools.add_tool(child_class)
    end

    def self.check?
      system "hash #{@name} 2>&-"
      $?.success?
    end
  end
end
